@@ -394,7 +394,10 @@ our %NODE_REQ;
    # monitoring
    mon0 => sub { # stop monitoring a port for another node
       my $portid = shift;
-      _unmonitor undef, $portid, delete $NODE{$SRCNODE}{rmon}{$portid};
+      # the if exists should not be needed, but there is apparently a bug
+      # elsewhere, and this works around that, silently suppressing that bug. sigh.
+      _unmonitor undef, $portid, delete $NODE{$SRCNODE}{rmon}{$portid}
+         if exists $NODE{$SRCNODE};
    },
    mon1 => sub { # start monitoring a port for another node
       my $portid = shift;
